# -*- coding: utf-8 -*-
# Copyright (C) 2019 fjord-technologies
# SPDX-License-Identifier: GPL-3.0-or-later
"""xtrargon_ng.classes.abstract.aws_boto3"""

import abc
import logging
import time
import os
import threading

import botocore
try:
    from botocore.hooks import AliasedEventEmitter as EventEmitter
except ImportError:
    from botocore.hooks import HierarchicalEmitter as EventEmitter

from awscli import clidriver, EnvironmentVariables
from boto3 import session

from sonicprobe import helpers

from six import iteritems

from dwho.classes.errors import DWhoConfigurationError


LOG = logging.getLogger('xtrargon_ng.classes.abstract.aws_boto3')


class XtrargonBotoNullEventHooks(botocore.hooks.BaseEventHooks):
    def _register(self, event_name, handler, unique_id=None, unique_id_uses_count=False): # pylint: disable-msg=unused-argument,no-self-use
        return

    def emit_until_response(self, event_name, **kwargs): # pylint: disable-msg=unused-argument,no-self-use
        return (None, None)


class XtrargonAwsBoto3Base(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, config = None, client_cfg = None):
        if not config or not isinstance(config, dict):
            config = {}

        self._config             = config
        self._client_cfg         = None
        self._botocore_sess      = None
        self._boto3_sess         = None

        self._client_thread      = {}
        self._client_thread_lock = threading.Lock()

        if 'client' in config and not client_cfg:
            self._client_cfg = botocore.client.Config(**config['client'])

    @staticmethod
    def create_driver():
        emitter  = EventEmitter()
        bsession = botocore.session.Session(EnvironmentVariables, event_hooks = emitter)
        clidriver._set_user_agent_for_session(bsession) # pylint: disable-msg=protected-access
        clidriver.load_plugins(bsession.full_config.get('plugins', {}),
                               event_hooks = emitter)
        return clidriver.CLIDriver(session = bsession).main

    def botocore_session(self, event_hooks = None):
        if self._botocore_sess:
            return self._botocore_sess

        self._botocore_sess = botocore.session.Session(include_builtin_handlers = False,
                                                       event_hooks = event_hooks)

        return self._botocore_sess

    def boto3_session(self, event_hooks = None):
        if self._boto3_sess:
            return self._boto3_sess

        self._boto3_sess = session.Session(
            botocore_session = self.botocore_session(event_hooks = event_hooks))

        return self._boto3_sess

    def botocore_client(self, service, event_hooks = None):
        sess = self.botocore_session(event_hooks = event_hooks)

        while True:
            try:
                return sess.create_client(service, config = self._client_cfg)
            except KeyError:
                time.sleep(0.2)

    def boto3_client(self, service, event_hooks = None):
        thread_ident = threading.currentThread().ident
        client_by_worker = bool(self._config.get('client_by_worker'))
        if client_by_worker:
            with self._client_thread_lock:
                if self._client_thread.get(thread_ident):
                    return self._client_thread[thread_ident]

        sess = self.boto3_session(
            self.botocore_session(event_hooks = event_hooks))

        while True:
            try:
                r = sess.client(service, config = self._client_cfg)
                if not client_by_worker:
                    return r

                with self._client_thread_lock:
                    self._client_thread[thread_ident] = r
                return r
            except KeyError:
                time.sleep(0.2)

    def client_config(self):
        return self._client_cfg

    def load_credentials(self, section, credentials = None):
        if not credentials and 'credentials' in self._config:
            credentials = self._config['credentials']

        if not credentials:
            return self

        cred = helpers.load_yaml_file(credentials)
        if not cred:
            raise DWhoConfigurationError("unable to load credentials")
        if not cred.get(section):
            raise DWhoConfigurationError("unable to find credentials for section: %r" % section)

        for k, v in iteritems(cred[section]):
            os.environ[k.upper()] = v

        return self

    def terminate(self):
        try:
            self._botocore_sess = None
            self._boto3_sess = None

            with self._client_thread_lock:
                for item in self._client_thread:
                    self._client_thread[item] = None
                self._client_thread.clear()
        except Exception:
            pass
